

# xml_set_attribv_occ
def test_xml_set_attribv_occ():
    from aiida_fleur.tools.xml_util import xml_set_attribv_occ
    pass

# xml_set_first_attribv
def test_xml_set_first_attribv():
    from aiida_fleur.tools.xml_util import xml_set_first_attribv
    pass

# xml_set_all_attribv
def test_xml_set_all_attribv():
    from aiida_fleur.tools.xml_util import xml_set_all_attribv
    pass

# xml_set_text
def test_xml_set_text():
    from aiida_fleur.tools.xml_util import xml_set_text
    pass

# xml_set_all_text
def test_xml_set_all_text():
    from aiida_fleur.tools.xml_util import xml_set_all_text
    pass

# create_tag
def test_create_tag():
    from aiida_fleur.tools.xml_util import create_tag
    pass

# delete_att
def test_delete_att():
    from aiida_fleur.tools.xml_util import delete_att
    pass

# delete_tag
def test_delete_tag():
    from aiida_fleur.tools.xml_util import delete_tag
    pass

# replace_tag
def test_replace_tag():
    from aiida_fleur.tools.xml_util import replace_tag
    pass

# set_species
def test_set_species():
    from aiida_fleur.tools.xml_util import set_species
    pass

# change_atomgr_att
def test_change_atomgr_att():
    from aiida_fleur.tools.xml_util import change_atomgr_att
    pass

# add_num_to_att
def test_add_num_to_att():
    from aiida_fleur.tools.xml_util import add_num_to_att
    pass

# eval_xpath
def test_eval_xpath():
    from aiida_fleur.tools.xml_util import eval_xpath
    pass

# eval_xpath2
def test_eval_xpath2():
    from aiida_fleur.tools.xml_util import eval_xpath2
    pass

# eval_xpath3
def test_eval_xpath3():
    from aiida_fleur.tools.xml_util import eval_xpath3
    pass

# get_xml_attribute
def test_get_xml_attribute():
    from aiida_fleur.tools.xml_util import get_xml_attribute
    pass

# get_inpxml_file_structure
# IMPORTANT: Here we need thats that tell us when the plugin has to be maintained, i.e Know thing in the inp schema where changed
# Is there a way to test for not yet know attributes? i.e if the plugin is complete? Back compatible?
# I.e make for each Fleur schema file, complete inp.xml file version a test if the attributes exists.

# make a test_class

    